# -*- coding: utf-8 -*-
"""
This script takes two bio files of annotation and prediction, and compute recall and precision for each NE label.
"""

import argparse
import logging
import os
import re

import editdistance
import edlib
import termtables as tt

THRESHOLD = 0.30
NOT_ENTITY_TAG = "O"


def parse_bio(path: str) -> dict:
    """Parse a BIO file to get text content, character-level NE tags and entity types count.

    Input : path to a valid BIO file
    Output format : { "words": str, "tags": list; "entity_count" : { tag : int} }
    """

    assert os.path.exists(path), f"Error: Input file {path} does not exist"

    words = []
    tags = []
    entity_count = {"All": 0}
    last_tag = None

    with open(path, "r") as fd:

        for line in list(filter(lambda x: x != "\n", fd.readlines())):

            word, label = line.split()

            # Preservation of '-' characters and avoid confusion with the dashes added later during the alignment
            word = word.replace("-", "§")
            words.append(word)

            try:
                tag = (
                    NOT_ENTITY_TAG
                    if label == NOT_ENTITY_TAG
                    else re.match(r"[BIES]-(.{3})", label)[1]
                )
            except TypeError:
                raise (
                    Exception(f"The file {path} given in input is not in BIO format.")
                )

            # Spaces will be added between words and have to get a tag
            # If previous word has the same tag as current, the space also gets the tag
            if last_tag is not None:
                if last_tag == tag:
                    tags.append(tag)
                else:
                    tags.append(NOT_ENTITY_TAG)

            # Add a tag for each letter in the word
            tags += [
                tag,
            ] * len(word)

            # Count nb entity for each type
            if (not label == "O") and (not last_tag == tag):
                entity_count[tag] = entity_count.get(tag, 0) + 1
                entity_count["All"] += 1

            last_tag = tag

        result = None

        if words:
            # Make string out of words
            result = dict()
            result["words"] = " ".join(words)
            result["tags"] = tags
            result["entity_count"] = entity_count
            assert len(result["words"]) == len(result["tags"])

    return result


def compute_matches(
    annotation: str, prediction: str, tags_annot: list, tags_predict: list
) -> dict:
    """Compute prediction score from annotation string to prediction string.

    Annotation and prediction strings should be the same length.

    For each entity in the annotation string, a match is found in the prediction.
    This is done in looking for a sub-string roughly at the same position in the prediction, and with the right entity-tag.
    Here is an example to illustrate the method used :

                   *-------*       *----*
    tags_annot   : PPPPPPPPPOOOOOOOCCCCCCOO
    annotation   : Tolkie-n- was a writer .
    prediction   : Tolkieene xas --writer .
    tags_predict : PPPPPPPPPOCCCCCCCCCCCCCC
                   *-------* <-----*----*->

    Each entity in the annotation string gets a prediction score based on the number
    of characters well predicted and labeled in the prediction string.
    The score of a label is the addition of entity scores divided by the number
    of entities.

    Inputs :
    annotation : str, example : "Tolkie-n- was a writer- -."
    prediction : str, example : "Tolkieene xas --writear ,."
    tags_annot : list of strings,   example : ['P','P','P','P','P','P','P','P','P','O', ...]
    tags_predict : list of string , example : ['P','P','P','P','P','P','P','P','P','O', ...]

    Output : {TAG1 : nb_entity_matched, ...}, example : {'All': 1, 'OCC': 0, 'PER': 1}
    """

    assert annotation
    assert prediction
    assert tags_annot
    assert tags_predict

    entity_count = {"All": 0}
    last_tag = NOT_ENTITY_TAG

    # Inspecting reference string
    for i, char_annot in enumerate(annotation):
        tag_ref = tags_annot[i]

        # If character not in entity
        if tag_ref == NOT_ENTITY_TAG:
            last_tag = NOT_ENTITY_TAG

        # Else, in entity
        else:

            # If beginning new entity
            if not tag_ref == last_tag:
                current_ref, current_compar = [], []
                last_tag = tag_ref
                found_aligned_beginning = False
                found_aligned_end = False

            current_ref.append(char_annot)

            # Searching character string corresponding with tag
            if not found_aligned_end and tags_predict[i] == tag_ref:

                # If just beginning new entity, backtrack tags on prediction string
                if len(current_ref) == 1:
                    j = i - 1
                    while j >= 0 and tags_predict[j] == tag_ref:
                        j -= 1
                    current_compar += prediction[j + 1 : i]

                found_aligned_beginning = True
                current_compar.append(prediction[i])

            # If tags don't match and beginning was found : end of predicted entity
            elif found_aligned_beginning:
                found_aligned_end = True

            # Detect end of entity in annotation
            if (i + 1 == len(annotation)) or (
                i + 1 < len(annotation) and not tags_annot[i + 1] == last_tag
            ):

                # Aligned entity may end further in prediction, so get the rest of the characters
                if not found_aligned_end:
                    j = i + 1
                    while j < len(tags_predict) and tags_predict[j] == tag_ref:
                        j += 1
                    for k in range(i + 1, j):
                        current_compar.append(prediction[k])

                # Normalize found character strings
                entity_ref = "".join(current_ref)
                entity_ref = entity_ref.replace("-", "")
                len_entity = len(entity_ref)
                entity_compar = "".join(current_compar)
                entity_compar = entity_compar.replace("-", "")

                # One entity is counted as recognized (score of 1) if the Levenhstein distance between the expected and predicted entities
                # represents less than 30% (Threshold) of the length of the expected entity.
                # Precision and recall will be computed for each category in comparing the numbers of recognized entities and expected entities
                score = (
                    1
                    if editdistance.eval(entity_ref, entity_compar) / len_entity
                    < THRESHOLD
                    else 0
                )
                entity_count[last_tag] = entity_count.get(last_tag, 0) + score
                entity_count["All"] += score

    return entity_count


def get_tags_aligned(original: str, aligned: str, tags_original: list) -> list:
    """Takes original string, original string tags and aligned string given by edlib.align.
    Returns a list of tags corresponding to the aligned string.

    Output format : list of strings
    """
    assert original
    assert aligned
    assert tags_original

    tags_aligned = []
    index_original = 0
    last_tag = NOT_ENTITY_TAG

    # Inspecting aligned string
    for i, char in enumerate(aligned):
        new_tag = ""

        # If original string has been fully processed, rest of tags are O ('-' characters at aligned end)
        if index_original >= len(original):
            new_tag = NOT_ENTITY_TAG

        # If current aligned char does not match current original char ('-' characters in aligned)
        # Keep last_tag and don't increment index_original
        elif not char == original[index_original]:
            new_tag = last_tag

        # Until matching of characters)
        else:
            new_tag = tags_original[index_original]
            last_tag = new_tag
            index_original += 1

        tags_aligned.append(new_tag)

    return tags_aligned


def compute_scores(
    annot_tags_count: dict, predict_tags_count: dict, matches: dict
) -> dict:
    """Compute Precision, Recall and F1 score for all entity types found in annotation and prediction.

    Each measure is given at document level, global score is a micro-average over tag types.

    Inputs :
    annot :   { TAG1(str) : nb_entity(int), ...}
    predict : { TAG1(str) : nb_entity(int), ...}
    matches : { TAG1(str) : nb_entity_matched(int), ...}

    Output :
    scores : { TAG1(str) : {"P" : float, "R" : float, "F1" : float}, ... }
    """

    annot_tags = set(annot_tags_count.keys())
    predict_tags = set(predict_tags_count.keys())
    tags = annot_tags | predict_tags

    scores = {tag: {"P": None, "R": None, "F1": None} for tag in tags}

    for tag in sorted(tags)[::-1]:
        nb_predict = predict_tags_count.get(tag)
        nb_annot = annot_tags_count.get(tag)
        nb_match = matches.get(tag, 0)
        prec = None if not nb_predict else nb_match / nb_predict
        rec = None if not nb_annot else nb_match / nb_annot
        f1 = (
            None
            if (prec is None) or (rec is None)
            else 0
            if (prec + rec == 0)
            else 2 * (prec * rec) / (prec + rec)
        )

        scores[tag]["predicted"] = nb_predict
        scores[tag]["matched"] = nb_match
        scores[tag]["P"] = prec
        scores[tag]["R"] = rec
        scores[tag]["F1"] = f1
        scores[tag]["Support"] = nb_annot

    return scores


def print_results(scores: dict):
    """Display final results.

    None values are kept to indicate the absence of a certain tag in either annotation or prediction.
    """
    header = ["tag", "predicted", "matched", "Precision", "Recall", "F1", "Support"]
    results = []
    for tag in sorted(scores.keys())[::-1]:
        prec = None if scores[tag]["P"] is None else round(scores[tag]["P"], 3)
        rec = None if scores[tag]["R"] is None else round(scores[tag]["R"], 3)
        f1 = None if scores[tag]["F1"] is None else round(scores[tag]["F1"], 3)

        results.append(
            [
                tag,
                scores[tag]["predicted"],
                scores[tag]["matched"],
                prec,
                rec,
                f1,
                scores[tag]["Support"],
            ]
        )
    tt.print(results, header, style=tt.styles.markdown)


def run(annotation: str, prediction: str) -> dict:
    """Compute recall and precision for each entity type found in annotation and/or prediction.

    Each measure is given at document level, global score is a micro-average over tag types.
    """

    # Get string and list of tags per character
    annot = parse_bio(annotation)
    predict = parse_bio(prediction)

    if not annot or not predict:
        raise Exception("No content found in annotation or prediction files.")

    # Align annotation and prediction
    align_result = edlib.align(annot["words"], predict["words"], task="path")
    nice_alignment = edlib.getNiceAlignment(
        align_result, annot["words"], predict["words"]
    )

    annot_aligned = nice_alignment["query_aligned"]
    predict_aligned = nice_alignment["target_aligned"]

    # Align tags from string alignment
    tags_annot_aligned = get_tags_aligned(annot["words"], annot_aligned, annot["tags"])
    tags_predict_aligned = get_tags_aligned(
        predict["words"], predict_aligned, predict["tags"]
    )

    # Get nb match
    matches = compute_matches(
        annot_aligned, predict_aligned, tags_annot_aligned, tags_predict_aligned
    )

    # Compute scores
    scores = compute_scores(annot["entity_count"], predict["entity_count"], matches)

    # Print results
    print_results(scores)

    return scores


def main():
    """Get arguments and run."""

    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser(description="Compute score of NER on predict.")
    parser.add_argument(
        "-a", "--annot", help="Annotation in BIO format.", required=True
    )
    parser.add_argument(
        "-p", "--predict", help="Prediction in BIO format.", required=True
    )
    args = parser.parse_args()

    run(args.annot, args.predict)


if __name__ == "__main__":
    main()
